#!/usr/bin/env python
"""
Create yml configuration files to submit R1->DL1 processing with EventDisplay

dl0_to_dl1_dark.yml
dl0_to_dl1_moon.yml

Then use:
cta-prod-submit <prodName> dl0_to_dl1_dark.yml
....

Usage example:
   cta-prod-create-workflow-evndisp-config <workflow_template.yml>

Example of workflow_template:

MCCampaign: PROD6
site: LaPalma
input_array_layout: Hyperarray
zenith: 60.0
analysis_prog_version: v5.13.2
compiler: gcc114_default
configuration_id: 16
pointing: North
"""

__RCSID__ = "$Id$"

from ruamel.yaml import YAML


from DIRAC.Core.Base.Script import Script

Script.parseCommandLine()

import DIRAC

yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)

arguments = Script.getPositionalArgs()
if len(arguments) != 1:
    Script.showHelp()

workflow_config_file = arguments[0]

with open(workflow_config_file) as stream:
    workflow_config = yaml.load(stream)

mandatory_keys = [
    "MCCampaign",
    "site",
    "input_array_layout",
    "zenith",
    "analysis_prog_version",
    "analysis_prog_compiler",
    "pointing",
]

for key in mandatory_keys:
    if key not in workflow_config:
        DIRAC.gLogger.notice(f"Missing mandatory key {key}")
        DIRAC.exit(-1)

MCCampaign = workflow_config["MCCampaign"]
site = workflow_config["site"]

if site not in ["LaPalma", "Paranal"]:
    DIRAC.gLogger.notice(f"Invalid value {site} for site")
    DIRAC.exit(-1)

input_array_layout = workflow_config["input_array_layout"]

pointing = workflow_config["pointing"]
if pointing not in ["North", "South"]:
    DIRAC.gLogger.notice(f"Invalid value {pointing} for pointing")
    DIRAC.exit(-1)

thetaP = workflow_config["zenith"]
zenith_str = str(int(thetaP)) + "deg"

version = workflow_config["analysis_prog_version"]
compiler = workflow_config["analysis_prog_compiler"]
configuration_id = workflow_config["configuration_id"]

common_dict = {}
common_dict["MCCampaign"] = MCCampaign
common_dict["configuration_id"] = configuration_id
common_dict["base_path"] = "/vo.cta.in2p3.fr/MC"


def create_dl0_dl1_config(moon):
    prod_dict = {}
    prod_step_list = []

    if moon == "dark":
        nsb = "NSB1x"
    elif moon == "half":
        nsb = "NSB5x"
    elif moon == "full":
        nsb = "NSB19x"

    id = 0
    for particle in [
        "gamma",
        "electron",
        "gamma-diffuse",
        "proton",
    ]:
        dataset = (
            MCCampaign.capitalize()
            + "_"
            + site
            + "_"
            + input_array_layout
            + "_"
            + nsb
            + "_"
            + particle
            + "_"
            + pointing
            + "_"
            + zenith_str
            + "_R1"
        )
        id += 1
        prod_step_dict = {}
        prod_step_dict["ID"] = id
        input_meta_query = {}
        input_meta_query["parentID"] = None
        input_meta_query["dataset"] = dataset
        prod_step_dict["input_meta_query"] = input_meta_query
        job_config = {}
        job_config["type"] = "EvndispProcessing"
        job_config["version"] = version
        job_config["compiler"] = compiler
        job_config["group_size"] = 5
        job_config["data_level"] = 1
        job_config["options"] = zenith_str + " " + moon
        job_config["tag"] = "production"
        job_config["catalogs"] = "DIRACFileCatalog"
        prod_step_dict["job_config"] = job_config
        prod_step_list.append(prod_step_dict)

    prod_dict["ProdSteps"] = prod_step_list
    prod_dict["Common"] = common_dict

    file_name = f"dl0_to_dl1_{moon}.yml"
    with open(file_name, "w") as file:
        yaml.dump(prod_dict, file)


@Script()
def main():
    create_dl0_dl1_config("dark")
    create_dl0_dl1_config("half")
    create_dl0_dl1_config("full")


########################################################
if __name__ == "__main__":
    main()



#!/usr/bin/env python

__RCSID__ = "$Id$"

# generic imports
import os

# DIRAC imports
from DIRAC.Core.Base.Script import Script
import DIRAC

Script.parseCommandLine(ignoreErrors=True)

from DIRAC import gLogger
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient


@Script()
def main():
    fcc = FileCatalogClient()
    dirac = Dirac()
    resJDL = dirac.getJobJDL(os.environ["JOBID"])

    idata = resJDL["Value"]["InputData"]
    if isinstance(idata, str):
        idata = []
        if "LFN" in resJDL["Value"]["InputData"]:
            idata.append(resJDL["Value"]["InputData"].split("LFN:")[1])
        else:
            idata.append(resJDL["Value"]["InputData"])
    else:
        idata = resJDL["Value"]["InputData"]

    for lfn in idata:
        gLogger.notice(f"Checking input file:\n {lfn} ")
        file_name = os.path.basename(lfn)
        code = os.system("./dirac_simtel_check " + file_name)
        if code != 0:
            gLogger.error(f"Corrupted file:\n {lfn}")
            gLogger.notice("Setting nsb=-1")
            res = fcc.setMetadata(lfn, {"nsb": -1})
            if not res["OK"]:
                return res

    DIRAC.exit()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Show the progress of DataProcessing transformations

Usage:
   cta-transformation-show-progress <transID or ascii file with a list of transID comma separated>

Example:
   cta-transformation-show-progress 2010
"""

__RCSID__ = "$Id$"

import os

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


def get_transformation_info(transID):
    transClient = TransformationClient()

    paramShowNames = [
        "TransformationID",
        "TransformationName",
        "Type",
        "Status",
        "Files_Total",
        "Files_PercentProcessed",
        "Files_Processed",
        "Files_Unused",
        "Jobs_TotalCreated",
        "Jobs_Waiting",
        "Jobs_Running",
        "Jobs_Done",
        "Jobs_Failed",
        "Jobs_Stalled",
    ]

    res = transClient.getTransformationSummaryWeb(
        {"TransformationID": transID}, [], 0, 1
    )

    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)

    if res["Value"]["TotalRecords"] > 0:
        paramNames = res["Value"]["ParameterNames"]
        for paramValues in res["Value"]["Records"]:
            paramShowValues = map(
                lambda pname: paramValues[paramNames.index(pname)], paramShowNames
            )
            showDict = dict(zip(paramShowNames, paramShowValues))
            files_PercentProcessed = showDict["Files_PercentProcessed"]
            transName = showDict["TransformationName"]

    return (transName, files_PercentProcessed)


#########################################################
@Script()
def main():
    Script.parseCommandLine()
    args = Script.getPositionalArgs()
    if len(args) != 1:
        Script.showHelp()

    # get arguments
    transIDs = []
    for arg in args[0].split(","):
        if os.path.exists(arg):
            lines = open(arg).readlines()
            for line in lines:
                for transID in line.split(","):
                    transIDs += [int(transID.strip())]
        else:
            transIDs.append(int(arg))

    values = []

    for transID in transIDs:
        transName, files_PercentProcessed = get_transformation_info(transID)
        values.append((transName, files_PercentProcessed))

    # print table
    gLogger.notice("\n|_. Transformation Name |_. Files Processed (%)| ")
    for value in values:
        gLogger.notice(f"|{'|'.join(map(str, value))}|")

    DIRAC.exit()


if __name__ == "__main__":
    main()



""" Collection of classes for DMS testing purposes
"""

__RCSID__ = "$Id$"

import DIRAC
from DIRAC import gLogger
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient

DIRAC.initialize()  # Initialize configuration


class ClientDMS:
    def __init__(self):
        self.dest_se1 = "CSCS-Disk"
        self.dest_se2 = "PIC-Disk"
        self.test_dir = "/vo.cta.in2p3.fr/tests/dms"
        self.test_file = "DMS_TestFile.txt"
        self.test_lfn = "/vo.cta.in2p3.fr/tests/dms/DMS_TestFile.txt"

    def write_test_file(self):
        with open(self.test_file, "w") as file:
            file.write("This is a test file")


class DMSmetadata:
    def __init__(
        self,
        dataset_name="Prod5b_LaPalma_AdvancedBaseline_NSB1x_electron_North_20deg_R1",
        dataset_dir="/vo.cta.in2p3.fr/datasets",
        dir_meta_dict=None,
        dir_meta_field_dict=None,
        file_meta_field_dict=None,
        file_meta_dict=None,
    ):
        self.dataset_name = dataset_name
        self.dataset_dir = dataset_dir
        self.file_list = []

        self.fc = FileCatalog()
        self.fcc = FileCatalogClient()
        var_char = "VARCHAR(128)"
        if not dir_meta_field_dict:
            self.dir_meta_field_dict = {
                "MCCampaign": var_char,
                "array_layout": var_char,
                "site": var_char,
                "particle": var_char,
                "thetaP": "float",
                "phiP": "float",
                "sct": var_char,
                "tel_sim_prog_version": var_char,
                "tel_sim_prog": var_char,
                "analysis_prog": var_char,
                "analysis_prog_version": var_char,
                "data_level": "int",
                "outputType": var_char,
                "configuration_id": "int",
                "merged": "int",
            }
        else:
            self.dir_meta_field_dict = dir_meta_field_dict

        if not file_meta_field_dict:
            self.file_meta_field_dict = {
                "nsb": "int",
                "div_ang": var_char,
                "split": var_char,
            }
        else:
            self.file_meta_field_dict = file_meta_field_dict

        if not dir_meta_dict:
            self.dir_meta_dict = {
                "MCCampaign": "Prod5bTest",
                "array_layout": "Advanced-Baseline",
                "site": "LaPalma",
                "particle": "electron",
                "thetaP": 20,
                "phiP": 180.0,
                "tel_sim_prog_version": "2020-06-29b",
                "tel_sim_prog": "sim_telarray",
                "data_level": -1,
                "outputType": "Data",
                "configuration_id": 15,
            }
        else:
            self.dir_meta_dict = dir_meta_dict

        if not file_meta_dict:
            self.file_meta_dict = {
                "nsb": 1,
            }
        else:
            self.file_meta_dict = file_meta_dict

    def create_directory(self):
        # Create directory where datasets will be registered
        gLogger.notice("Create directory where datasets will be registered")
        res = self.fc.createDirectory(self.dataset_dir)
        if not res["OK"]:
            gLogger.error(f"ERROR: Failed to create directory {self.dataset_dir}")
        return res

    def remove_dir_md_from_dfc(self):
        # Remove directory metadata fields from the DFC if present
        gLogger.notice("Remove directory metadata fields from the DFC if present")
        for meta_field in self.dir_meta_field_dict:
            res = self.fc.deleteMetadataField(meta_field)
            if not res["OK"]:
                gLogger.error(f"ERROR: Failed to delete metadata field {meta_field}")
            return res

    def remove_file_md_from_dfc(self):
        # Remove file metadata fields from the DFC if present
        gLogger.notice("Remove file metadata fields from the DFC if present")
        for meta_field in self.dir_meta_field_dict:
            res = self.fc.deleteMetadataField(meta_field)
            if not res["OK"]:
                gLogger.error(f"ERROR: Failed to delete metadata field {meta_field}")
            return res

    def add_dir_md_in_dfc(self):
        # Add directory metadata fields in DFC
        gLogger.notice("Adding directory metadata fields in DFC")
        for key, value in self.dir_meta_field_dict.items():
            res = self.fc.addMetadataField(key, value)
            if not res["OK"]:
                gLogger.error(f"ERROR: Failed to add metadata field {key}")
            return res

    def add_file_md_in_dfc(self):
        # Add file metadata fields in DFC
        gLogger.notice("Adding file metadata fields in DFC")
        for key, value in self.file_meta_field_dict.items():
            res = self.fc.addMetadataField(key, value, metaType="-f")
            if not res["OK"]:
                gLogger.error(f"ERROR: Failed to add metadata field {key}")
            return res

    def main(self):
        suite = {
            "create_directory": self.create_directory,
            "add_dir_md_in_dfc": self.add_dir_md_in_dfc,
            "add_file_md_in_dfc": self.add_file_md_in_dfc,
        }

        return suite



#!/usr/bin/env python

__RCSID__ = "$Id$"

# generic imports
from multiprocessing import Pool

# DIRAC imports
from DIRAC import gLogger
from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file
from DIRAC.Core.Base.Script import Script

Script.setUsageMessage(
    """
Bulk removal of a list of files from the catalog
Usage:
   cta-prod-unregister-file <ascii file with lfn list>
"""
)

Script.parseCommandLine(ignoreErrors=True)
# Must comes after parseCommandLine
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

fc = FileCatalog()


def unregisterFile(lfn):
    res = fc.removeFile(lfn)
    if res["OK"]:
        if "Failed" in res["Value"]:
            if lfn in res["Value"]["Failed"]:
                gLogger.error(res["Value"]["Failed"][lfn])
            elif lfn in res["Value"]["Successful"]:
                gLogger.notice("Successfully removed from the catalog", lfn)
            else:
                gLogger.error("Unexpected error result", res["Value"])
        else:
            gLogger.notice("Successfully removed from the catalog", lfn)
    else:
        gLogger.error("Failed to remove file from the catalog:", res["Message"])

    if not res["OK"]:
        gLogger.error(res["Message"])


@Script()
def main():
    args = Script.getPositionalArgs()
    if len(args) > 0:
        infile = args[0]
    else:
        Script.showHelp()

    infileList = read_inputs_from_file(infile)
    p = Pool(10)
    p.map(unregisterFile, infileList)


if __name__ == "__main__":
    main()



import time
from typing import Any, Union
from rich.console import Console
from rich.table import Table

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.ReturnValues import DOKReturnType, DErrorReturnType
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.TransformationSystem.Client.Transformation import (
    Transformation,
)
from DIRAC.TransformationSystem.Client.TransformationClient import (
    TransformationClient,
)
from DIRAC.TransformationSystem.Agent.TransformationAgentsUtilities import (
    TransformationAgentsUtilities,
)
from DIRAC.ProductionSystem.Client.ProductionClient import ProductionClient
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from CTADIRAC.Core.Utilities.return_values import DReportReturnType, s_report


class FailoverUtilities(TransformationAgentsUtilities):
    """
    Some useful methods used by TransformationFailoverAgent
    """

    def __init__(self, agent_name) -> None:
        TransformationAgentsUtilities.__init__(self)
        self.notif_client = NotificationClient()
        self.trans_client = TransformationClient()
        self.prod_client = ProductionClient()

        self.mail_from = None
        self.mail_to = None
        self.enabled = False

        self.attributes_of_interest: list[str] = [
            "Status",
            "MinorStatus",
            "ApplicationStatus",
            "Site",
        ]

    def get_eligible_transformation(
        self, status, type_list
    ) -> Union[Any, DOKReturnType]:
        """Select transformations of given status and type.
        using the TransformationClient"""
        cond_dict = {"Status": status, "Type": type_list}
        if not type_list:
            cond_dict = {"Status": status}
        res = self.trans_client.getTransformations(condDict=cond_dict)
        if not res["OK"]:
            return res
        transformations = {}
        for trans_info in res["Value"]:
            trans_id = trans_info["TransformationID"]
            transformations[str(trans_id)] = trans_info
        return S_OK(transformations)

    def get_transformation_tasks(self, trans_id: int) -> Union[Any, DOKReturnType]:
        """get transformations tasks
        using the TransformationClient"""
        res = self.trans_client.getTransformationTasks({"TransformationID": trans_id})
        if not res["OK"]:
            gLogger.error(res["Message"])

        if len(res["Value"]) == 0:
            gLogger.notice(f"[{trans_id}] No tasks selected")

        return res

    def get_transformation_files(self, cond_dict) -> Union[Any, list]:
        """Get the transformation files
        using the TransformationClient"""

        res = self.trans_client.getTransformationFiles(cond_dict)
        return res

    def is_transformation_processed(self, trans_id: int) -> bool:
        res_trans_stat = self.trans_client.getTransformationStats(trans_id)
        is_processed = False
        if res_trans_stat["OK"]:
            try:
                if (
                    res_trans_stat["Value"]["Processed"]
                    == res_trans_stat["Value"]["Total"]
                ):
                    is_processed = True
            except KeyError:
                gLogger.info(f"[{trans_id}] Transformation got no 'Processed' files.")
        else:
            gLogger.error(
                f"[{trans_id}] Can't getTransformationStats: {res_trans_stat}"
            )
        return is_processed

    def complete_transformation(
        self, trans_id: int
    ) -> Union[DReportReturnType, DErrorReturnType]:
        res_complete = self.trans_client.completeTransformation(trans_id)
        if res_complete["OK"]:
            return s_report(False, f"[{trans_id}] Transformation Completed.")
        else:
            return S_ERROR(
                f"[{trans_id}]: Can't complete transformation: {res_complete}"
            )

    def get_parent_transformation_id(self, trans_id: int) -> Union[int, None]:
        res = self.prod_client.get_parent_transformation(trans_id)
        if res["OK"]:
            if res["Value"]:
                try:
                    parent_id = res["Value"][0][0]
                    if isinstance(parent_id, int):
                        return parent_id
                except (TypeError, IndexError):
                    gLogger.error(
                        f"[{trans_id}] Error while getting parent transformation: {res}"
                    )
                    return None
            else:
                gLogger.info(f"[{trans_id}] Parent ID has empty value")
                return None
        else:
            gLogger.error(
                f"[{trans_id}] Error while getting parent transformation: {res}"
            )
            return None

    def get_parent_type(self, parent_id: int) -> Union[Any, None]:
        res = self.trans_client.getTransformations(
            condDict={"TransformationID": parent_id}
        )
        if res["OK"]:
            return res["Value"][0]["Type"]

    def set_task_status(self, trans_id: int, task_id: int, status: str):
        transformation = Transformation(transID=trans_id, transClient=self.trans_client)
        return transformation.setTaskStatus(taskID=task_id, status=status)

    def sorting_key(self, index1, index2):
        return lambda item: (item[index1], item[index2])

    def create_mail_message(
        self, transformation_reports: dict[str, dict[str, str]], html=False
    ):
        if html:
            return self.create_html_mail_message(transformation_reports)
        else:
            return self.create_text_mail_message(transformation_reports)

    def create_text_mail_message(
        self, transformation_reports: dict[str, dict[str, str]]
    ) -> str:
        """Sort the transformation by types
        and create the text string message"""
        message = ""
        for trans_type in transformation_reports:
            message += f"Report on {trans_type} transformations:"
            for report in transformation_reports[trans_type].values():
                message += report + "\n"
        return message

    def create_html_mail_message(
        self, transformation_reports: dict[str, dict[str, str]]
    ) -> str:
        """Sort the transformation by types
        and create the html string message"""

        message = ""
        for trans_type in transformation_reports:
            message += """<!DOCTYPE html>
            <html>
            <head>
            <meta charset="UTF-8">
            <style>
            .r1 {font-style: italic}
            .r2 {font-weight: bold}
            .r3 {color: #008000; text-decoration-color: #008000; font-weight: bold}
            .r4 {color: #800000; text-decoration-color: #800000; font-weight: bold}
            body {
                color: #000000;
                background-color: #ffffff;
            }
            </style>
            </head>
            <body>
                <pre style="font-family:Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace"><code><span class="r1">"""
            message += f"Report on {trans_type} transformations:"
            message += """</span>
            </code></pre>
            </body>
            </html>"""
            for report in transformation_reports[trans_type].values():
                message += report + "\n"
        return message

    def init_rich_table(self, title: str, fields: list[str]) -> Table:
        """Init Rich Table:"""
        table = Table(title=title, style="bold")
        for f in fields:
            table.add_column(f)
        return table

    def generate_table(self, table: Table, html=False) -> str:
        """Export rich table"""
        console = Console(record=True, width=200)
        console.print(table)
        if html:
            html_output: str = console.export_html()
        else:
            html_output: str = console.export_text()
        return html_output

    def create_transformation_jobs_report(
        self,
        trans_id: int,
        count_status_type: dict,
        count_minor_status: dict,
        count_status: dict,
        total_nb_jobs: int,
    ) -> str:
        """Create the Rich transformation jobs report"""
        title: str = f"Transformation: {trans_id}"
        fields: list[str] = self.attributes_of_interest + ["Total"]
        table: Table = self.init_rich_table(title, fields)

        self.add_rows_in_table(
            count_minor_status,
            count_status,
            count_status_type,
            total_nb_jobs,
            table,
        )

        return self.generate_table(table, self.html)

    def add_rows_in_table(
        self,
        count_minor_status: dict,
        count_status: dict,
        count_status_type: dict,
        total_nb_jobs: int,
        table: Table,
    ) -> None:
        """Add rows in the table"""
        next_minor_status = ""
        next_status = ""
        # Sort the StatusType by Status then by MinorStatus:
        count_status_type_sorted: list = sorted(
            count_status_type,
            key=lambda x: (
                x[self.attributes_of_interest.index("Status")],
                x[self.attributes_of_interest.index("MinorStatus")],
            ),
        )
        N: int = len(count_status_type_sorted)
        n = 0
        for status_tuple in count_status_type_sorted:
            row = list(status_tuple) + [str(count_status_type[status_tuple])]
            table.add_row(*row)
            # Add Total
            if n + 1 <= N:
                if n + 1 < N:
                    next_status, next_minor_status = self.get_next_status(
                        n, count_status_type_sorted
                    )
                # Add minor_status total
                if self.should_add_minor_status_total(
                    n, next_minor_status, status_tuple, N
                ):
                    row = self.minor_status_total_row(
                        row, count_minor_status, status_tuple, count_status
                    )
                    table.add_row(*row, style="bold yellow")
                # Add status total
                if self.should_add_status_total(n, next_status, status_tuple, N):
                    row, style = self.status_total_row(
                        row, count_status, status_tuple, total_nb_jobs
                    )
                    table.add_row(*row, style=style)
            n += 1
        final_row: list[str] = self.final_row(total_nb_jobs)
        table.add_row(*final_row, style="bold")

    def get_next_status(
        self, n: int, count_status_type_sorted: list
    ) -> tuple[Any, Any]:
        """Get the next job status"""
        next_status: str = (
            count_status_type_sorted[n + 1][0]
            if n + 1 < len(count_status_type_sorted)
            else ""
        )
        next_minor_status: str = (
            count_status_type_sorted[n + 1][1]
            if n + 1 < len(count_status_type_sorted)
            else ""
        )
        return next_status, next_minor_status

    def should_add_minor_status_total(
        self, n: int, next_minor_status: str, status_tuple: tuple, tot: int
    ) -> Union[Any, bool]:
        """Condition to add the total of minor status"""
        return (next_minor_status != status_tuple[1] and next_minor_status != "") or (
            n + 1 == tot
        )

    def minor_status_total_row(
        self,
        row: list,
        count_minor_status: dict,
        status_tuple: tuple,
        count_status: dict,
    ) -> list[str]:
        """Create the minor status total row"""
        row[1] = f"Total: {row[1]}"
        row[-1] = (
            f"{count_minor_status[status_tuple[1]]} "
            f"({round(count_minor_status[status_tuple[1]]/count_status[status_tuple[0]]*100, 1)}%)"
        )
        row[-2] = "-"
        row[-3] = "-"
        return row

    def should_add_status_total(
        self, n: int, next_status: str, status_tuple: tuple, tot: int
    ) -> Union[Any, bool, str]:
        """Condition on creating the total status"""
        return (next_status != status_tuple[0] and next_status) or (n + 1 == tot)

    def status_total_row(
        self, row: list, count_status: dict, status_tuple: tuple, total_nb_jobs: int
    ) -> tuple[list, str]:
        """Create the status total row"""
        row[0] = f"Total: {row[0]}"
        row[-1] = (
            f"{count_status[status_tuple[0]]} "
            f"({round(count_status[status_tuple[0]]/total_nb_jobs*100, 1)}%)"
        )
        row[1] = "-"
        if any(status in row[0] for status in [JobStatus.FAILED, JobStatus.KILLED]):
            style = "bold red"
        elif any(status in row[0] for status in [JobStatus.DONE, JobStatus.COMPLETED]):
            style = "bold green"
        else:
            style = "bold orange3"
        return row, style

    def final_row(self, total_nb_jobs: int) -> list[str]:
        """Create the final row with total number of jobs"""
        final_row: list[str] = [
            "Total # tasks:",
            "-",
            "-",
            "-",
            str(total_nb_jobs),
        ]
        return final_row

    def create_reassigned_lfn_report(
        self, trans_id: int, reassigned_lfns: dict[int, list]
    ) -> str:
        """Create report on reassigned files"""
        n_lfns: int = sum(len(lfns) for lfns in reassigned_lfns.values())
        table: Table = self.init_rich_table(
            title=f"Reassigned {n_lfns} LFNs on transformation {trans_id}:",
            fields=["taskID", "LFN"],
        )
        for taskid, lfns_list in reassigned_lfns.items():
            for lfn in lfns_list:
                table.add_row(*[str(taskid), str(lfn)])
        return self.generate_table(table)

    def send_mail(self, subject: str, message: str, html=False) -> None:
        gLogger.info(f"Sending mail from {self.mail_from} to {self.mail_to}")
        res = self.notif_client.sendMail(
            self.mail_to, subject, message, self.mail_from, localAttempt=True, html=html
        )
        if res["OK"]:
            gLogger.info(f"Report succesfully sent to {self.mail_to}")
        else:
            gLogger.error(res)

    @classmethod
    def measure_time(cls, func):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = func(*args, **kwargs)
            end_time = time.time()
            elapsed_time = end_time - start_time
            gLogger.info(f"{func.__name__} took {elapsed_time:.2f} seconds to execute")
            return result

        return wrapper



""" Manage data and meta-data
"""

__RCSID__ = "$Id$"

# generic imports
import os
import glob
import json
import collections

# DIRAC imports
import DIRAC
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.TransformationSystem.Client import TransformationFilesStatus
from DIRAC.Core.Utilities import List
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite
from DIRAC.Interfaces.API.Dirac import Dirac


class ProdDataManager:
    """Manage data and meta-data"""

    def __init__(self, catalogs=["DIRACFileCatalog"]):
        """Constructor"""
        self.setupCatalogClient(catalogs)
        self.printCatalogConfig(catalogs)
        self.setupDataManagerClient(catalogs)
        self.transClient = TransformationClient()
        # ## Get the TransformationID
        self.TransformationID = "0000"  # used for local execution or submission to wms
        self.InputData = []
        if "JOBID" in os.environ:
            jobID = os.environ["JOBID"]
            dirac = Dirac()
            res = dirac.getJobJDL(jobID)
            if "TransformationID" in res["Value"]:
                self.TransformationID = res["Value"]["TransformationID"]
            if "InputData" in res["Value"]:
                self.InputData = res["Value"]["InputData"]

    def setupCatalogClient(self, catalogs):
        """Init FileCatalog client
        Ideally we would like to use only FileCatalog but it doesn't work for setMetadata
        because the returned value has a wrong format. So use of FileCatalogClient instead
        """
        self.fc = FileCatalog(catalogs)
        self.fcc = FileCatalogClient()

    def printCatalogConfig(self, catalogs):
        """Dumps catalog configuration"""
        for catalog in catalogs:
            res = self.fc._getCatalogConfigDetails(catalog)
            DIRAC.gLogger.debug("CatalogConfigDetails:", res)

    def setupDataManagerClient(self, catalogs):
        """Init DataManager client"""
        self.dm = DataManager(catalogs)

    def _getSEList(self, SEType="ProductionOutputs", DataType="SimtelProd"):
        """get from CS the list of available SE for data upload"""
        opsHelper = Operations()
        optionName = os.path.join(SEType, DataType)
        SEList = opsHelper.getValue(optionName, [])
        SEList = List.randomize(SEList)
        DIRAC.gLogger.notice(f"List of {SEType} SE: {SEList} ")

        # # Check if the local SE is in the list. If yes try it first by reversing list order
        localSEList = []
        res = getSEsForSite(DIRAC.siteName())
        if res["OK"]:
            localSEList = res["Value"]

        retainedlocalSEList = []
        for localSE in localSEList:
            if localSE in SEList:
                DIRAC.gLogger.notice(
                    "The local Storage Element is an available SE: ", localSE
                )
                retainedlocalSEList.append(localSE)
                SEList.remove(localSE)

        SEList = retainedlocalSEList + SEList
        if len(SEList) == 0:
            return DIRAC.S_ERROR("Error in building SEList")

        return DIRAC.S_OK(SEList)

    def _checkemptydir(self, path):
        """check that the directory is not empty"""
        if len(glob.glob(path)) == 0:
            error = f"Empty directory: {path}"
            return DIRAC.S_ERROR(error)
        else:
            return DIRAC.S_OK()

    def _getRunPath(self, filemetadata):
        """format path to string with 1 digit precision
        run_number is taken from filemetadata
        filemetadata can be a dict or the run_number itself
        """
        if isinstance(filemetadata, dict):
            run_number = int(filemetadata["runNumber"])
        else:
            run_number = int(filemetadata)
        run_numberMod = run_number % 1000
        runpath = "%03dxxx" % ((run_number - run_numberMod) / 1000)
        return runpath

    def _formatPath(self, path):
        """format path to string with 1 digit precision"""
        if isinstance(path, float) or isinstance(path, int):
            path = f"{path:.1f}"
        return str(path)

    def createMDTopDirectory(self, metadata, basepath, program_category):
        """create meta data structure"""

        ### Create the directory structure
        md = json.loads(metadata, object_pairs_hook=collections.OrderedDict)

        path = basepath
        process_program = program_category + "_prog"
        for key, value in collections.OrderedDict(
            (k, md[k])
            for k in ("MCCampaign", "site", "particle", process_program)
            if k in md
        ).items():
            path = os.path.join(path, self._formatPath(value))
            res = self.fc.createDirectory(path)
            if not res["OK"]:
                return res

            # Set directory metadata for each subdir: 'site', 'particle', 'process_program'
            res = self.fcc.setMetadata(path, {key: value})
            if not res["OK"]:
                return res

        return DIRAC.S_OK(path)

    def createMDSubDirectory(
        self, metadata, base_path, program_category, output_type_list, trans_id
    ):
        """create meta data structure"""

        ### Create the directory structure
        md = json.loads(metadata, object_pairs_hook=collections.OrderedDict)
        process_program = program_category + "_prog"

        # Create the TransformationID subdir and set MD
        path = os.path.join(base_path, str(trans_id))
        res = self.fc.createDirectory(path)
        if not res["OK"]:
            return res

        process_program_version = process_program + "_version"
        res = self.fcc.setMetadata(
            path,
            {
                k: md[k]
                for k in (
                    "phiP",
                    "thetaP",
                    "array_layout",
                    process_program_version,
                    "sct",
                )
                if k in md
            },
        )
        if not res["OK"]:
            return res

        # Create subdir according to the output_type (Data, Log, Model) and set MD
        transformation_path = path
        for output_type in output_type_list:
            path = os.path.join(transformation_path, output_type)
            res = self.fc.createDirectory(path)
            if not res["OK"]:
                return res

            res = self.fcc.setMetadata(path, {"outputType": output_type})
            if not res["OK"]:
                return res

            # Set additional metadata at subdir level if not already defined
            for metadataKey in [
                "data_level",
                "configuration_id",
            ]:
                res = self.fcc.setMetadata(path, {metadataKey: md[metadataKey]})
                if not res["OK"]:
                    return res

            if output_type in ["Data", "Log"]:
                res = self.fcc.setMetadata(path, {"merged": md.get("merged", 0)})
                if not res["OK"]:
                    return res

        return DIRAC.S_OK(transformation_path)

    def getTransformationPath(self, metadata, basepath, program_category):
        """get the transformation path"""

        ### Create the directory structure
        md = json.loads(metadata, object_pairs_hook=collections.OrderedDict)

        path = basepath
        process_program = program_category + "_prog"
        for key, value in collections.OrderedDict(
            (k, md[k])
            for k in ("MCCampaign", "site", "particle", process_program)
            if k in md
        ).items():
            path = os.path.join(path, self._formatPath(value))

        # get the TransformationID subdir
        transformation_path = os.path.join(path, self.TransformationID)

        return DIRAC.S_OK(transformation_path)

    def putAndRegister(self, lfn, localfile, filemetadata, DataType):
        """put and register one file and set metadata"""
        # ## Get the list of Production SE
        res = self._getSEList("ProductionOutputs", DataType)
        if res["OK"]:
            ProductionSEList = res["Value"]
        else:
            return res

        # ## Get the list of Failover SE
        res = self._getSEList("ProductionOutputsFailover", DataType)
        if res["OK"]:
            FailoverSEList = res["Value"]
        else:
            return res

        # ## Upload file to a Production SE
        res = self._putAndRegisterToSEList(lfn, localfile, ProductionSEList)
        if not res["OK"]:
            DIRAC.gLogger.error(
                f"Failed to upload file to any Production SE: {ProductionSEList}"
            )
            # ## Upload file to a Failover SE
            res = self._putAndRegisterToSEList(lfn, localfile, FailoverSEList)
            if not res["OK"]:
                return DIRAC.S_ERROR(
                    f"Failed to upload file to any Failover SE: {FailoverSEList}"
                )

        ### Set file metadata also adding jobID metadata
        if res["OK"]:
            fmd = json.loads(filemetadata)
            if "JOBID" in os.environ:
                fmd.update({"jobID": os.environ["JOBID"]})
            res = self.fc.setMetadata(lfn, fmd)
            if not res["OK"]:
                return res

            return DIRAC.S_OK()

    def _putAndRegisterToSEList(self, lfn, localfile, SEList):
        """put and register one file to one SE in the SEList"""
        # Try to upload file to a SE in the list
        for SE in SEList:
            msg = f"Try to upload local file: {localfile} \nwith LFN: {lfn} \nto {SE}"
            DIRAC.gLogger.notice(msg)
            res = self.dm.putAndRegister(lfn, localfile, SE)
            DIRAC.gLogger.notice(res)
            # Check if failed
            if not res["OK"]:
                DIRAC.gLogger.error(f"Failed to putAndRegister {lfn}")
            else:
                return DIRAC.S_OK()
        return DIRAC.S_ERROR()

    def setTransformationFileStatus(self, lfn, status):
        if status == "PROCESSED":
            res = self.transClient.setFileStatusForTransformation(
                self.TransformationID, TransformationFilesStatus.PROCESSED, lfn
            )
        elif status == "PROBLEMATIC":
            res = self.transClient.setFileStatusForTransformation(
                self.TransformationID, TransformationFilesStatus.PROBLEMATIC, lfn
            )
        elif status == "UNUSED":
            res = self.transClient.setFileStatusForTransformation(
                self.TransformationID, TransformationFilesStatus.UNUSED, lfn
            )
        else:
            DIRAC.gLogger.error(f"Not allowed File Status {status}")
            return DIRAC.S_ERROR()

        return res

    def cleanLocalFiles(self, datadir, pattern="*"):
        """remove files matching pattern in datadir"""

        for localfile in glob.glob(os.path.join(datadir, pattern)):
            DIRAC.gLogger.notice("Removing local file: ", localfile)
            os.remove(localfile)

        return DIRAC.S_OK()
